# flake8: noqa
try:
    from molotov.api import (
        scenario,
        setup,
        global_setup,
        teardown,
        global_teardown,
        setup_session,
        teardown_session,
        scenario_picker,
        events,
    )
    from molotov.util import request, json_request
    from molotov.util import set_var, get_var
    from molotov.session import get_context
except ImportError:
    pass  # first import

__version__ = "2.4"
